/*
 * exc-abort.cpp: Gisle Vanem <gvanem@yahoo.no> 2011
 */
#include "exc-abort.h"
#include "StackWalker.h"

bool use_debug     = false;
bool no_fatal_exit = false;

/* Specialized stackwalker-output class; prints to console */

class StackWalkerToConsole : public StackWalker
{
protected:
  virtual void OnSymInit (LPCSTR szSearchPath, DWORD symOptions, LPCSTR szUserName)
  {
    if (use_debug)
       printf ("  Symbol search-path: \"%s\"\n"
               "  User-name: \"%s\"\n", szSearchPath, szUserName);
  }

  virtual void OnLoadModule (LPCSTR img, LPCSTR mod, DWORD64 baseAddr, DWORD size,
                             DWORD result, LPCSTR symType, LPCSTR pdbName, ULONGLONG fileVersion)
  {
    if (use_debug)
       StackWalker::OnLoadModule (img, mod, baseAddr, size,
                                  result, symType, pdbName, fileVersion);
  }

  virtual void OnOutput (LPCSTR szText)
  {
    if (use_debug || strncmp(szText, "ERROR: ", 7))  /* drop error lines unless 'use_debug==1' */
       printf ("  %s", szText);
  }
};


static bool filter_set = false;
static void (*orig_abort_handler)(int) = NULL;

static void new_abort_handler (int sig);
static BOOL PreventSetUnhandledExceptionFilter (void);
static LONG __stdcall CrashHandlerExceptionFilter (EXCEPTION_POINTERS* pExPtrs);

#if defined(STANDALONE)
  /*
   * If this file is included from test.cpp, remove DllMain() from linkage.
   */
  BOOL APIENTRY DllMain_bogus (HANDLE, DWORD, LPVOID);
  #define DllMain DllMain_bogus

  int init_exc_abort (void)
  {
    return DllMain_bogus (NULL, DLL_PROCESS_ATTACH, NULL);
  }
#endif

BOOL APIENTRY DllMain (HANDLE hModule,
                       DWORD  reason_for_call,
                       LPVOID lpReserved)
{
  if (reason_for_call == DLL_PROCESS_ATTACH &&
      !IsDebuggerPresent() && !filter_set)
  {
    const char *env = getenv ("EXC-ABORT");

    if (env && strstr(env, "DEBUG"))
       use_debug = true;
    if (env && strstr(env, "NO_FATAL"))
       no_fatal_exit = true;

    /* Set new signal-handler for trapping SIGABRT signals
     */
    orig_abort_handler = signal (SIGABRT, new_abort_handler);

#ifdef _MSC_VER
    _set_abort_behavior (0, _WRITE_ABORT_MSG);
#endif

    /* set global exception handler (for handling all unhandled exceptions)
     */
    SetUnhandledExceptionFilter (CrashHandlerExceptionFilter);
    filter_set = true;
  }

  (void) hModule;
  (void) lpReserved;
  return (TRUE);
}

static void new_abort_handler (int sig)
{
  /** \to-do: show a better MessageBox(). Show backtrace
   * from where abort() was called; skip the 2 first CRT locations (raise, abort).
   */
  printf ("Got SIGABRT. Backtrace:\n");
  StackWalkerToConsole sw;
  sw.ShowCallstack();

  if (orig_abort_handler && orig_abort_handler != SIG_DFL)
    (*orig_abort_handler) (sig);

  _exit (-1);
}

// For more info about "PreventSetUnhandledExceptionFilter" see:
// "SetUnhandledExceptionFilter" and VC8
// http://blog.kalmbachnet.de/?postid=75
// and "Unhandled exceptions in VC8 and above for x86 and x64"
// http://blog.kalmbach-software.de/2008/04/02/unhandled-exceptions-in-vc8-and-above-for-x86-and-x64/

#if defined(_M_X64) || defined(_M_IX86)
LPTOP_LEVEL_EXCEPTION_FILTER WINAPI
MyDummySetUnhandledExceptionFilter (LPTOP_LEVEL_EXCEPTION_FILTER lpTopLevelExceptionFilter)
{
  return (NULL);
}

static BOOL PreventSetUnhandledExceptionFilter (void)
{
  HMODULE hKernel32 = LoadLibrary ("kernel32.dll");
  void   *pOrgEntry;

  if (!hKernel32)
     return FALSE;

  pOrgEntry = (void*) GetProcAddress (hKernel32, "SetUnhandledExceptionFilter");
  if (!pOrgEntry)
     return FALSE;

  DWORD  dwOldProtect = 0;
  SIZE_T jmpSize = 5;

#if defined(_M_X64) || defined(__x86_64__)
  jmpSize = 13;
#endif

  BOOL bProt = VirtualProtect (pOrgEntry, jmpSize,
                               PAGE_EXECUTE_READWRITE, &dwOldProtect);
  BYTE newJump[20];
  void *pNewFunc = (void*)&MyDummySetUnhandledExceptionFilter;

#if defined(_M_IX86)
  DWORD dwOrgEntryAddr = (DWORD) pOrgEntry;
  dwOrgEntryAddr += jmpSize; // add 5 for 5 op-codes for jmp rel32
  DWORD dwNewEntryAddr = (DWORD) pNewFunc;
  DWORD dwRelativeAddr = dwNewEntryAddr - dwOrgEntryAddr;
  // JMP rel32: Jump near, relative, displacement relative to next instruction.
  newJump[0] = 0xE9;  // JMP rel32
  memcpy (&newJump[1], &dwRelativeAddr, sizeof(pNewFunc));
#elif defined(_M_X64)
  newJump[0] = 0x49;  // MOV R15, ...
  newJump[1] = 0xBF;  // ...
  memcpy (&newJump[2], &pNewFunc, sizeof (pNewFunc));
  // pCur += sizeof (ULONG_PTR);
  newJump[10] = 0x41;  // JMP R15, ...
  newJump[11] = 0xFF;  // ...
  newJump[12] = 0xE7;  // ...
#endif

  SIZE_T bytesWritten;
  BOOL   bRet = WriteProcessMemory (GetCurrentProcess(),
                                    pOrgEntry, newJump, jmpSize,
                                    &bytesWritten);
  if (bProt != FALSE)
  {
    DWORD dwBuf;
    VirtualProtect (pOrgEntry, jmpSize, dwOldProtect, &dwBuf);
  }
  return bRet;
}
#else
#error "This code works only for x86 and x64!"
#endif

static LONG __stdcall CrashHandlerExceptionFilter (EXCEPTION_POINTERS* pExPtrs)
{
  if (pExPtrs->ExceptionRecord->ExceptionCode == EXCEPTION_STACK_OVERFLOW)
  {
    static char MyStack[1024*128];  // be sure that we have enought space...

#if defined(__GNUC__)

#if 0 // f*ck this now
    __asm__ __volatile__ (
           "movl %1, %%eax\n"
           "movl %%eax, %%esp"
        :: "r" (MyStack));
#endif

#elif defined(__clang__)

#elif defined(_M_IX86)
    // it assumes that DS and SS are the same!!! (this is the case for Win32)
    // change the stack only if the selectors are the same (this is the case for Win32)
    //__asm push offset MyStack[1024*128];
    //__asm pop esp;
  __asm mov eax,offset MyStack[1024*128];
  __asm mov esp,eax;
#endif  /* __GNUC__ */
  }

  StackWalkerToConsole sw;  // output to console
  sw.ShowCallstack (GetCurrentThread(), pExPtrs->ContextRecord);

  char lString [500];
  _snprintf_s (lString, sizeof(lString),
               "*** Unhandled Exception!\r\n"
               " Code:    0x%08lX\r\n"
               " Flags:   %lu\r\n"
               " Address: 0x%" PRIxPTR "\r\n",
               (DWORD)pExPtrs->ExceptionRecord->ExceptionCode,
               (DWORD)pExPtrs->ExceptionRecord->ExceptionFlags,
               (UINT_PTR)pExPtrs->ExceptionRecord->ExceptionAddress);

  if (no_fatal_exit)
       _exit (-1);
  else FatalAppExit (-1, lString);

  return (EXCEPTION_CONTINUE_SEARCH);
}
